
#pragma once

#include <dataspace/capability.h>
#include <base/signal.h>
#include <base/rpc.h>
#include <session/session.h>
#include <packet_stream_tx/packet_stream_tx.h>
#include <packet_stream_rx/packet_stream_rx.h>

namespace Tee {

	using Genode::Packet_stream_sink;
	using Genode::Packet_stream_source;
	using Genode::size_t;

	using Session_id = Genode::uint64_t;

	class Packet_descriptor;
	struct Session;

	enum class Result {
		NO_ERROR,
		UNAVAILABLE,
		EXECUTION_ERROR,
		EXECUTION_TIMEOUT,
		INVALID_CLIENT_CERTIFICATE,
		INVALID_SIGNATURE,
	};
}

/**
 * TEE packet type
 */
class Tee::Packet_descriptor : public Genode::Packet_descriptor
{
	private:
		Result     _result = Result::NO_ERROR;
		size_t     _length = 0;
		Session_id _session_id = 0;

	public:
		Packet_descriptor()
		: Genode::Packet_descriptor(0, 0) { }

		Packet_descriptor(const Genode::off_t offset, const Genode::size_t size)
		: Genode::Packet_descriptor(offset, size) { }

		/**
		 * The length of the actual data to transmit
		 */
		Genode::size_t length() const { return _length; }
		void length(const Genode::size_t length) { _length = length; }

		Result result() const { return _result; }
		void result(const Result result) { _result = result; }
		bool ok() const { return _result == Result::NO_ERROR; }

		Session_id session_id() const { return _session_id; }
		void session_id(const Session_id id) { _session_id = id; }
};

/*
 * TEE session interface
 *
 * A TEE session can be used to transmit and receive protobuf packets with a
 * specific header. Payload is communicated over the packet-stream interface
 * set up between 'Session_client' and 'Session_server'.
 *
 * Even though the methods 'tx', 'tx_channel', 'rx', and 'rx_channel' are
 * specific for the client side of the TEE session interface, they are part of
 * the abstract 'Session' class to enable the client-side use of the TEE
 * interface via a pointer to the abstract 'Session' class. This way, we can
 * transparently co-locate the packet-stream server with the client in same
 * program.
 */
struct Tee::Session : Genode::Session
{
	enum { QUEUE_SIZE = 1024 };

	/*
	 * Types used by the client stub code and server implementation
	 *
	 * The acknowledgement queue has always the same size as the submit
	 * queue. We access the packet content as a char pointer.
	 */
	typedef Genode::Packet_stream_policy<Tee::Packet_descriptor,
	                                     QUEUE_SIZE, QUEUE_SIZE, char> Policy;

	typedef Packet_stream_tx::Channel<Policy> Tx;
	typedef Packet_stream_rx::Channel<Policy> Rx;

	/**
	 * \noapi
	 */
	static const char *service_name() { return "Tee"; }

	/*
	 * A TEE session consumes a dataspace capability for the server-side
	 * session object, a session capability, two packet-stream dataspaces for
	 * rx and tx, and four signal context capabilities for the data-flow
	 * signals.
	 */
	enum { CAP_QUOTA = 8 };

	virtual ~Session() { }

	/**
	 * Request packet-transmission channel
	 */
	virtual Tx *tx_channel() { return 0; }

	/**
	 * Request packet-reception channel
	 */
	virtual Rx *rx_channel() { return 0; }

	/**
	 * Request client-side packet-stream interface of tx channel
	 */
	virtual Tx::Source *tx() { return 0; }

	/**
	 * Request client-side packet-stream interface of rx channel
	 */
	virtual Rx::Sink *rx() { return 0; }

	/*******************
	 ** RPC interface **
	 *******************/

	GENODE_RPC(Rpc_tx_cap, Genode::Capability<Tx>, _tx_cap);
	GENODE_RPC(Rpc_rx_cap, Genode::Capability<Rx>, _rx_cap);

	GENODE_RPC_INTERFACE(Rpc_tx_cap, Rpc_rx_cap);
};
